""" Code to contain example and other empty classes mainly for tests
"""

# I am making a make system so that unit tests in EASE can load fake targets that will stay constant for the tests
# Due to its wider applicability im putting these here

from .astroclasses import Planet, Star, Binary, System, Parameters, PlanetParameters, StarParameters, BinaryParameters
from . import astroclasses as ac


def genExampleSystem():
    systemPar = Parameters()
    systemPar.addParam('name', 'Example System ' + str(ac._ExampleSystemCount))
    systemPar.addParam('distance', 58)
    systemPar.addParam('declination', '+04 05 06')
    systemPar.addParam('rightascension', '01 02 03')

    exampleSystem = System(systemPar.params)
    exampleSystem.flags.addFlag('Fake')

    ac._ExampleSystemCount += 1

    return exampleSystem


def genExampleBinary():
    binaryPar = BinaryParameters()
    binaryPar.addParam('name', 'Example Binary {0}AB'.format(ac._ExampleSystemCount))
    binaryPar.addParam('semimajoraxis', 10)  # TODO add realistic value
    binaryPar.addParam('period', 10)  # TODO add realistic value
    # TODO add the rest of binary parameters

    exampleBinary = Binary(binaryPar.params)
    exampleBinary.flags.addFlag('Fake')

    # generate other star
    exampleStar2 = genExampleStar('B', False)
    exampleStar2.parent = exampleBinary
    exampleBinary._addChild(exampleStar2)

    exampleSystem = genExampleSystem()
    exampleSystem._addChild(exampleBinary)
    exampleBinary.parent = exampleSystem

    return exampleBinary


def genExampleStar(binaryLetter='', heirarchy=True):
    """ generates example star, if binaryLetter is true creates a parent binary object, if heirarchy is true will create a
    system and link everything up
    """

    starPar = StarParameters()
    starPar.addParam('age', '7.6')
    starPar.addParam('magB', '9.8')
    starPar.addParam('magH', '7.4')
    starPar.addParam('magI', '7.6')
    starPar.addParam('magJ', '7.5')
    starPar.addParam('magK', '7.3')
    starPar.addParam('magV', '9.0')
    starPar.addParam('mass', '0.98')
    starPar.addParam('metallicity', '0.43')
    starPar.addParam('name', 'Example Star {0}{1}'.format(ac._ExampleSystemCount, binaryLetter))
    starPar.addParam('name', 'HD {0}{1}'.format(ac._ExampleSystemCount, binaryLetter))
    starPar.addParam('radius', '0.95')
    starPar.addParam('spectraltype', 'G5')
    starPar.addParam('temperature', '5370')

    exampleStar = Star(starPar.params)
    exampleStar.flags.addFlag('Fake')

    if heirarchy:
        if binaryLetter:
            exampleBinary = genExampleBinary()
            exampleBinary._addChild(exampleStar)
            exampleStar.parent = exampleBinary
        else:
            exampleSystem = genExampleSystem()
            exampleSystem._addChild(exampleStar)
            exampleStar.parent = exampleSystem

    return exampleStar


def genExamplePlanet(binaryLetter=''):
    """ Creates a fake planet with some defaults
    :param `binaryLetter`: host star is part of a binary with letter binaryletter
    :return:
    """

    planetPar = PlanetParameters()
    planetPar.addParam('discoverymethod', 'transit')
    planetPar.addParam('discoveryyear', '2001')
    planetPar.addParam('eccentricity', '0.09')
    planetPar.addParam('inclination', '89.2')
    planetPar.addParam('lastupdate', '12/12/08')
    planetPar.addParam('mass', '3.9')
    planetPar.addParam('name', 'Example Star {0}{1} b'.format(ac._ExampleSystemCount, binaryLetter))
    planetPar.addParam('period', '111.2')
    planetPar.addParam('radius', '0.92')
    planetPar.addParam('semimajoraxis', '0.449')
    planetPar.addParam('temperature', '339.6')
    planetPar.addParam('transittime', '2454876.344')
    planetPar.addParam('separation', '330', {'unit': 'AU'})

    examplePlanet = Planet(planetPar.params)
    examplePlanet.flags.addFlag('Fake')

    exampleStar = genExampleStar(binaryLetter=binaryLetter)
    exampleStar._addChild(examplePlanet)
    examplePlanet.parent = exampleStar

    return examplePlanet


examplePlanet = genExamplePlanet()
exampleStar = examplePlanet.parent
exampleSystem = exampleStar.parent



""" This module contains some plotting functions and plot types for easy plot creation
"""
import os
import sys
import math
from collections import OrderedDict

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import rcParams
import time

from . import astroquantities as aq
from . import astroclasses as ac


rcParams.update({'figure.autolayout': True})


class _GlobalFigure(object):
    """ sets up the figure and subfigure object with all the global parameters.
    """

    def __init__(self, size='small'):
        self.setup_fig(size)

    def setup_fig(self, size='small'):
        self.set_size(size)

    def set_size(self, size):
        """ choose a preset size for the plot
        :param size: 'small' for documents or 'large' for presentations
        """
        if size == 'small':
            self._set_size_small()
        elif size == 'large':
            self._set_size_large()
        else:
            raise ValueError('Size must be large or small')

    def _set_size_small(self):
        self.fig = plt.figure(figsize=(5, 4))
        self.ax = self.fig.add_subplot(1, 1, 1)
        self.set_title_size(10)
        self.set_axis_label_size(12)
        self.set_axis_tick_label_size(12)

    def _set_size_large(self):
        self.fig = plt.figure(figsize=(10, 7.5))
        self.ax = self.fig.add_subplot(1, 1, 1)
        self.set_title_size(20)
        self.set_axis_label_size(20)
        self.set_axis_tick_label_size(20)

    def set_global_font_size(self, fontsize):
        ax = self.ax
        for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] + ax.get_xticklabels() + ax.get_yticklabels()):
            item.set_fontsize(fontsize)
        plt.draw()

    def set_title_size(self, fontsize):
        self.ax.title.set_fontsize(fontsize)
        plt.draw()

    def set_axis_label_size(self, fontsize):
        for axis in (self.ax.xaxis.label, self.ax.yaxis.label):
            axis.set_fontsize(fontsize)
        plt.draw()

    def set_axis_tick_label_size(self, fontsize):
        for axis in self.ax.get_xticklabels() + self.ax.get_yticklabels():
            axis.set_fontsize(fontsize)
        plt.draw()

    # set_foregroundcolor / set_backgroundcolor from Jasonmc https://gist.github.com/jasonmc/1160951
    def set_foregroundcolor(self, color):
         '''For the specified axes, sets the color of the frame, major ticks,
             tick labels, axis labels, title and legend
         '''

         ax = self.ax

         for tl in ax.get_xticklines() + ax.get_yticklines():
             tl.set_color(color)
         for spine in ax.spines:
             ax.spines[spine].set_edgecolor(color)
         for tick in ax.xaxis.get_major_ticks():
             tick.label1.set_color(color)
         for tick in ax.yaxis.get_major_ticks():
             tick.label1.set_color(color)
         ax.axes.xaxis.label.set_color(color)
         ax.axes.yaxis.label.set_color(color)
         ax.axes.xaxis.get_offset_text().set_color(color)
         ax.axes.yaxis.get_offset_text().set_color(color)
         ax.axes.title.set_color(color)
         lh = ax.get_legend()
         if lh != None:
             lh.get_title().set_color(color)
             lh.legendPatch.set_edgecolor('none')
             labels = lh.get_texts()
             for lab in labels:
                 lab.set_color(color)
         for tl in ax.get_xticklabels():
             tl.set_color(color)
         for tl in ax.get_yticklabels():
             tl.set_color(color)
         plt.draw()

    def set_backgroundcolor(self, color):
         '''Sets the background color of the current axes (and legend).
             Use 'None' (with quotes) for transparent. To get transparent
             background on saved figures, use:
             pp.savefig("fig1.svg", transparent=True)
         '''
         ax = self.ax
         ax.patch.set_facecolor(color)
         lh = ax.get_legend()
         if lh != None:
             lh.legendPatch.set_facecolor(color)

         plt.draw()

    def set_y_axis_log(self, logscale=True):
        if logscale:
            self.ax.set_yscale('log')
        else:
            self.ax.set_yscale('linear')

    def set_x_axis_log(self, logscale=True):
        if logscale:
            self.ax.set_xscale('log')
        else:
            self.ax.set_xscale('linear')


class _AstroObjectFigs(_GlobalFigure):
    """ contains extra functions for dealing with input of astro objects
    """

    def __init__(self, objectList, size='small'):
        _GlobalFigure.__init__(self, size)
        self.objectList = objectList  # list of planets, stars etc
        self._objectType = self._getInputObjectTypes()  # are we dealing with planets, stars etc

    def _getInputObjectTypes(self):
        # get the type of the first object
        firstObject = self.objectList[0]
        firstObjectType = type(firstObject)

        for astroObject in self.objectList:
            if not firstObjectType == type(astroObject):
                raise TypeError('Input object list contains mixed types ({0} and {1})'.format(firstObjectType,
                                                                                              type(astroObject)))
        return firstObjectType

    def _getParLabelAndUnit(self, param):
        """ checks param to see if it contains a parent link (ie star.) then returns the correct unit and label for the
         job from the parDicts
        :return:
        """

        firstObject = self.objectList[0]

        if isinstance(firstObject, ac.Planet):
            if 'star.' in param:
                return _starPars[param[5:]]  # cut off star. part
            else:
                return _planetPars[param]
        elif isinstance(firstObject, ac.Star):
            return _starPars[param]
        else:
            raise TypeError('Only Planets and Star object are currently supported, you gave {0}'.format(type(firstObject)))

    # OECPy specific functions
    def _gen_label(self, param, unit):
        # TODO could be star or other in future
        parValues = self._getParLabelAndUnit(param)

        if unit is None:
            return parValues[0]
        else:
            unitsymbol = self._get_unit_symbol(unit)
            return '{0} ({1})'.format(parValues[0], unitsymbol)

    def _get_unit_symbol(self, unit):
        """ Every quantities object has a symbol, but i have added `latex_symbol` for some types. This first checks for
        a latex symbol and then if it fails will just use symbol
        :return:
        """
        try:
            return '${0}$'.format(unit.latex_symbol)
        except AttributeError:
            return unit.symbol


class _BaseDataPerClass(_AstroObjectFigs):  # hangover from ETLOS's multiple child plot classes
    """ Base class for plots counting the results by a attribute. Child classes must modify
    * _classVariables (self._allowedKeys, )
    * _getSortKey (take the planet, turn it into a key)
    """

    def __init__(self, astroObjectList, unit=None, size='small'):  # added unit here as class will break without it anyway
        _AstroObjectFigs.__init__(self, astroObjectList, size)

        self._classVariables()  # add info from child classes
        self.unit = unit
        self.resultsByClass = self._processResults()

    def _set_size_small(self):
        _AstroObjectFigs._set_size_small(self)
        self.xticksize = 8

    def _set_size_large(self):
        _AstroObjectFigs._set_size_large(self)
        self.xticksize = 15

    def _classVariables(self):
        """ Variables to be loaded in init by child classes

        must set
        *self._allowedKeys = (tuple of keys)
        """
        self._allowedKeys = ('Class')

    def _getSortKey(self, planet):
        """ Takes a planet and turns it into a key to be sorted by
        :param planet:
        :return:
        """

        return 'Class'

    def _processResults(self):
        """ Checks each result can meet SNR requirments, adds to count
        :return:
        """

        resultsByClass = self._genEmptyResults()

        for astroObject in self.objectList:
            sortKey = self._getSortKey(astroObject)
            resultsByClass[sortKey] += 1

        return resultsByClass

    def _getPlotData(self):
        """ Turns the resultsByClass Dict into a list of bin groups skipping the uncertain group if empty

        return: (label list, ydata list)
        :rtype: tuple(list(str), list(float))
        """
        resultsByClass = self.resultsByClass

        try:
            if resultsByClass['Uncertain'] == 0:  # remove uncertain tag if present and = 0
                resultsByClass.pop('Uncertain', None)
        except KeyError:
            pass

        plotData = list(zip(*resultsByClass.items()))  # (labels, ydata)

        return plotData

    def plotBarChart(self, title='', xlabel=None, c='#3ea0e4', label_rotation=False):
        ax = self.ax

        labels, ydata = self._getPlotData()

        numItems = float(len(ydata))

        ind = np.arange(numItems)  # the x locations for the groups
        ind /= numItems  # between 0 and 1

        spacePerBar = 1./numItems
        gapratio = 0.5  # gap to bar ratio, 0.5 is even
        width = spacePerBar*gapratio
        gap = spacePerBar - width

        ax.bar(ind, ydata, width, color=c)
        ax.set_xticklabels(labels)
        ax.set_xticks(ind+(width/2.))

        for axis in self.ax.get_xticklabels():
            axis.set_fontsize(self.xticksize)

        if self.unit is None:  # NOTE this is hacked in so it only works with DataPerParameterClass
            self.unit = self._getParLabelAndUnit(self._planetProperty)[1]  # use the default unit defined in this class
        self._yaxis_unit = self.unit

        if xlabel is None:
            plt.xlabel(self._gen_label(self._planetProperty, self.unit))
        else:
            plt.xlabel(xlabel)

        if label_rotation:
            plt.xticks(rotation=label_rotation)
        plt.ylabel('Number of Planets')  # TODO could be stars, binaries etc
        plt.title(title)
        plt.xlim([min(ind)-gap, max(ind)+(gap*2)])
        plt.draw()

    def plotPieChart(self, title=None, cmap_name='Pastel2'):
        if sys.hexversion >= 0x02700000:
            self.fig.set_tight_layout(True)

        # Generate plot data
        labels, ydata = self._getPlotData()
        totalobjects = float(np.sum(ydata))
        fracs = [ynum / totalobjects for ynum in ydata]

        explode = np.zeros(len(fracs))  # non zero makes the slices come out of the pie

        # plot pie chart
        cmap = plt.cm.get_cmap(cmap_name)
        colors = cmap(np.linspace(0., 0.9, len(fracs)))
        plt.pie(fracs, explode=explode, labels=labels, autopct='%1.1f%%', shadow=False, startangle=90,
                colors=colors)

        # generate plot / axis labels
        for axis in self.ax.get_xticklabels():
            axis.set_fontsize(self.xticksize)

        if self.unit is None:  # NOTE this is hacked in so it only works with DataPerParameterClass
            self.unit = self._getParLabelAndUnit(self._planetProperty)[1]  # use the default unit defined in this class
        self._yaxis_unit = self.unit

        if title is None:
            title = 'Planet {0} Bins'.format(self._gen_label(self._planetProperty, self.unit))  # NOTE not always planet

        plt.title(title)
        plt.xlim(-1.5, 1.5)

    def saveAllBarChart(self, filepath, *args, **kwargs):
        self.plotBarChart(*args, **kwargs)
        plt.savefig(os.path.join(filepath))

    def _genEmptyResults(self):
        """ Uses allowed keys to generate a empty dict to start counting from
        :return:
        """

        allowedKeys = self._allowedKeys

        keysDict = OrderedDict()  # Note: list comprehension take 0 then 2 then 1 then 3 etc for some reason. we want strict order
        for k in allowedKeys:
            keysDict[k] = 0


        resultsByClass = keysDict

        return resultsByClass


class DataPerParameterBin(_BaseDataPerClass):
    """ Generates Data for planets per parameter bin"""

    def __init__(self, results, planetProperty, binLimits, unit=None, size='small'):
        """
        :param planetProperty: property of planet to bin. IE 'e' for eccentricity, 'star.magV' for magV
        :param binLimits: list of bin limits (lower limit, upper, upper, maximum) (note you can have maximum +)
        :param unit: unit to scale param to (see general plotter)
        :return:
        """

        self._binlimits = binLimits
        self._planetProperty = planetProperty

        self._genKeysBins()  # Generate the bin keys/labels (must do before base class processes results)
        _BaseDataPerClass.__init__(self, results, unit, size)

    def _getSortKey(self, planet):
        """ Takes a planet and turns it into a key to be sorted by
        :param planet:
        :return:
        """

        value = eval('planet.'+self._planetProperty)

        # TODO some sort of data validation, either before or using try except

        if self.unit is not None:
            try:
                value = value.rescale(self.unit)
            except AttributeError:  # either nan or unitless
                pass

        return _sortValueIntoGroup(self._allowedKeys[:-1], self._binlimits, value)

    def _classVariables(self):
        pass  # Overload as we dont want it to set anything in this class

    def _genKeysBins(self):
        """ Generates keys from bins, sets self._allowedKeys normally set in _classVariables
        """
        binlimits = self._binlimits

        allowedKeys = []
        midbinlimits = binlimits

        if binlimits[0] == -float('inf'):
            midbinlimits = binlimits[1:]  # remove the bottom limit
            allowedKeys.append('<{0}'.format(midbinlimits[0]))

        if binlimits[-1] == float('inf'):
            midbinlimits = midbinlimits[:-1]

        lastbin = midbinlimits[0]

        for binlimit in midbinlimits[1:]:
            if lastbin == binlimit:
                allowedKeys.append('{0}'.format(binlimit))
            else:
                allowedKeys.append('{0} to {1}'.format(lastbin, binlimit))
            lastbin = binlimit

        if binlimits[-1] == float('inf'):
            allowedKeys.append('{0}+'.format(binlimits[-2]))

        allowedKeys.append('Uncertain')
        self._allowedKeys = allowedKeys


class GeneralPlotter(_AstroObjectFigs):
    """ This class should be able to create a plot with lots of options like the online visual plots. In future it
    should be turned into a GUI
    """

    def __init__(self, objectList, xaxis=None, yaxis=None, xunit=None, yunit=None, xaxislog=False, yaxislog=False, size='small'):
        """
        :param objectList: list of astro objects to use in plot ie planets, stars etc
        :param xaxis: value to use on the xaxis, should be a variable or function of the objects in objectList. ie 'R'
            for the radius variable and 'calcDensity()' for the calcDensity function
        :param yaxis: value to use on the yaxis, should be a variable or function of the objects in objectList. ie 'R'
            for the radius variable and 'calcDensity()' for the calcDensity function

        :type objectList: list, tuple
        :type xaxis: str
        :type yaxis: str
        """
        _AstroObjectFigs.__init__(self, objectList, size)

        # setup vars - to be replaced by themes
        self.set_marker_color()
        self.set_marker_size()

        # params that can be modified by user
        self.marker = 'o'

        # set later
        self.xlabel = None
        self.ylabel = None

        # Handle given parameters
        if xaxis:
            self.set_xaxis(xaxis, xunit)
        else:
            self._xaxis = None

        if yaxis:
            self.set_yaxis(yaxis, yunit)
        else:
            self._yaxis = None

        self.set_x_axis_log(xaxislog)
        self.set_y_axis_log(yaxislog)

    def _set_size_large(self):
        _AstroObjectFigs._set_size_large(self)
        self.set_marker_size(60)

    def plot(self):
        xaxis = [float(x) for x in self._xaxis]
        yaxis = [float(y) for y in self._yaxis]

        assert(len(xaxis) == len(yaxis))

        plt.scatter(xaxis, yaxis, marker=self.marker, facecolor=self._marker_color, edgecolor=self._edge_color,
                    s=self._marker_size)

        plt.xlabel(self.xlabel)
        plt.ylabel(self.ylabel)

    def set_xaxis(self, param, unit=None, label=None):
        """ Sets the value of use on the x axis
        :param param: value to use on the xaxis, should be a variable or function of the objects in objectList. ie 'R'
        for the radius variable and 'calcDensity()' for the calcDensity function

        :param unit: the unit to scale the values to, None will use the default
        :type unit: quantities unit or None

        :param label: axis label to use, if None "Parameter (Unit)" is generated here and used
        :type label: str
        """

        if unit is None:
            unit = self._getParLabelAndUnit(param)[1]  # use the default unit defined in this class
        self._xaxis_unit = unit

        self._xaxis = self._set_axis(param, unit)
        if label is None:
            self.xlabel = self._gen_label(param, unit)
        else:
            self.xlabel = label

    def set_yaxis(self, param, unit=None, label=None):
        """ Sets the value of use on the yaxis
        :param param: value to use on the yaxis, should be a variable or function of the objects in objectList. ie 'R'
        for the radius variable and 'calcDensity()' for the calcDensity function

        :param unit: the unit to scale the values to
        :type unit: quantities unit or None

        :param label: axis label to use, if None "Parameter (Unit)" is generated here and used
        :type label: str
        """
        if unit is None:
            unit = self._getParLabelAndUnit(param)[1]  # use the default unit defined in this class
        self._yaxis_unit = unit

        self._yaxis = self._set_axis(param, unit)
        if label is None:
            self.ylabel = self._gen_label(param, unit)
        else:
            self.ylabel = label

    def _set_axis(self, param, unit):
        """ this should take a variable or a function and turn it into a list by evaluating on each planet
        """
        axisValues = []
        for astroObject in self.objectList:
            try:
                value = eval('astroObject.{0}'.format(param))
            except ac.HierarchyError:  # ie trying to call planet.star and one planet is a lone ranger
                value = np.nan

            if unit is None:  # no unit to rescale (a aq.unitless quanitity would otherwise fail with ValueError)
                axisValues.append(value)
            else:
                try:
                    axisValues.append(value.rescale(unit))
                except AttributeError:  # either nan or unitless
                    axisValues.append(value)

        return axisValues

    def set_marker_color(self, color='#3ea0e4', edgecolor='k'):
        """ set the marker color used in the plot
        :param color: matplotlib color (ie 'r', '#000000')
        """
        # TODO allow a colour set per another variable
        self._marker_color = color
        self._edge_color = edgecolor

    def set_marker_size(self, size=30):
        """ set the marker size used in the plot
        :param size: matplotlib size
        """
        self._marker_size = size


class DiscoveryMethodByYear(object):

    def __init__(self, planet_list, methods_to_plot=('RV', 'transit', 'Other'), skip_solar_system_planets=True):
        """ Produces a labelled stacked bar chart of planet discovery methods


        :param planet_list: list of planet objects
        :param methods_to_plot: list of discovery methods to plot (OEC syntax)
        , use 'other' to group non specified methods
        :param skip_solar_system_planets: don't include solar system planets
        (including them will really stretch the xscale)
        """
        self.planet_list = planet_list
        self.skip_solar_system_planets = skip_solar_system_planets
        self.methods_to_plot = methods_to_plot

    def setup_keys(self):
        """ Build the initial data dictionary to store the values
        """

        discovery_methods = {}
        discovery_years = {}
        nan_list = []

        # Initial Loop to get keys
        for planet in self.planet_list:
            if 'Solar System' in planet.params['list'] and self.skip_solar_system_planets:
                continue
            try:
                discovery_methods[planet.discoveryMethod] += 1
            except KeyError:
                discovery_methods[planet.discoveryMethod] = 1

            try:
                discovery_years[planet.discoveryYear] += 1
            except KeyError:
                discovery_years[planet.discoveryYear] = 1

            if planet.discoveryMethod is np.nan:
                nan_list.append(planet)

        self.nan_list = nan_list

        return discovery_years

    def generate_data(self):

        discovery_years = self.setup_keys()

        _year_list = np.array([year for year in discovery_years.keys() if year is not np.nan])
        year_list = range(_year_list.min(), _year_list.max()+1)  # +1 needed as its < not <=

        discovery_methods = {k:0 for k in self.methods_to_plot}
        plot_matrix = {k:discovery_methods.copy() for k in year_list}

        # TODO (ryan) the same sort of  loop as setup_keys again? really?
        for planet in self.planet_list:
            if 'Solar System' in planet.params['list'] and self.skip_solar_system_planets:
                continue

            if planet.discoveryMethod in self.methods_to_plot:
                discovery_method = planet.discoveryMethod
            else:
                discovery_method = 'Other'

            try:
                plot_matrix[planet.discoveryYear][discovery_method] += 1
            except KeyError:  # nan year
                pass

        return plot_matrix, year_list, discovery_years

    def plot(self, method_labels=None, exodata_credit=True):
        """
        :param method_labels: list of legend labels for the methods, i.e. give
        'Radial Velocity' rather than 'RV'

        :return:
        """

        if method_labels is None:
            method_labels = self.methods_to_plot

        plot_matrix, year_list, discovery_years = self.generate_data()

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)

        ind = np.arange(len(year_list))

        # output from seaborn.color_palette("Set1", n_colors=8)
        colors = [(0.89411765336990356, 0.10196078568696976, 0.10980392247438431),
                  (0.21602460800432691, 0.49487120380588606, 0.71987698697576341),
                  (0.30426760128900115, 0.68329106055054012, 0.29293349969620797),
                  (0.60083047361934883, 0.30814303335021526, 0.63169552298153153),
                  (1.0, 0.50591311045721465, 0.0031372549487095253),
                  (0.99315647868549117, 0.9870049982678657, 0.19915417450315812),
                  (0.65845446095747107, 0.34122261685483596, 0.1707958535236471),
                  (0.95850826852461868, 0.50846600392285513, 0.74492888871361229)]

        width = 0.9
        bottom = np.zeros_like(year_list)
        for i, method in enumerate(self.methods_to_plot):
            bar_data = [plot_matrix[year][method] for year in year_list]
            plt.bar(ind, bar_data, width, color=colors[i], bottom=bottom, label=method_labels[i], linewidth=0)

            bottom += np.array(bar_data) # plot the next bar stacked

        plt.ylabel('Number of planets', fontsize=14)
        plt.xlabel('Year', fontsize=14)
        plt.xticks(ind+width/2., year_list, rotation=70, fontsize=13)
        plt.xlim(0, ind[-1]+1)
        plt.legend(loc=2, fontsize=13)

        if exodata_credit is True:
            ax.text(2, 420, 'Generated by ExoData on {}'.format(time.strftime("%d/%m/%Y")), fontsize=11)

        for i, year in enumerate(year_list):
            try:
                total_planets = discovery_years[year]
            except KeyError:  # year with no planets discovered
                continue  # Dont label
            y_pos = bottom[i]+10
            x_pos = i

            ax.text(x_pos, y_pos, '{}'.format(total_planets), fontsize=9)

        plt.grid(b=True, which='both', color='0.9', linestyle='-')

        return fig


def _sortValueIntoGroup(groupKeys, groupLimits, value):
    """ returns the Key of the group a value belongs to
    :param groupKeys: a list/tuple of keys ie ['1-3', '3-5', '5-8', '8-10', '10+']
    :param groupLimits: a list of the limits for the group [1,3,5,8,10,float('inf')] note the first value is an absolute
    minimum and the last an absolute maximum. You can therefore use float('inf')
    :param value:
    :return:
    """

    if not len(groupKeys) == len(groupLimits)-1:
        raise ValueError('len(groupKeys) must equal len(grouplimits)-1 got \nkeys:{0} \nlimits:{1}'.format(groupKeys,
                                                                                                         groupLimits))

    if math.isnan(value):
        return 'Uncertain'

    # TODO add to other if bad value or outside limits
    keyIndex = None

    if value == groupLimits[0]:  # if value is == minimum skip the comparison
        keyIndex = 1
    elif value == groupLimits[-1]:  # if value is == minimum skip the comparison
        keyIndex = len(groupLimits)-1
    else:
        for i, limit in enumerate(groupLimits):
            if value < limit:
                keyIndex = i
                break

    if keyIndex == 0:  # below the minimum
        raise BelowLimitsError('Value {0} below limit {1}'.format(value, groupLimits[0]))

    if keyIndex is None:
        raise AboveLimitsError('Value {0} above limit {1}'.format(value, groupLimits[-1]))

    return groupKeys[keyIndex-1]


class OutOfLimitsError(Exception):
    pass


class BelowLimitsError(Exception):
    pass


class AboveLimitsError(Exception):
    pass

# A parameter dict based on the current astroclass list and the parameter selected. Here the standard
# units, full name for labels and short name as stored. All values that could be used are listed here and many are
# commented because i dont think they should be used or will not work without further code to interpret
# These dicts allow axis labels for the plots
_planetPars = {
    # paramKey    : (axis label,      unit)
    #'isTransiting': ('Is Transiting', bool), TODO add bool and grouped values
    'calcTransitDuration()': ('Transit Duration', aq.day),
    'calcTransitDepth()': ('Transit Depth', None),
    #'type()': ('Planet Class', None),
    # 'massType()': ('Planet Mass Class', None),
    # 'radiusType()': ('Planet Radius Class', None),
    # 'temptype()': ('Planet Temp Class', None),
    'mu': ('Mean Molecular Weight', None),
    'albedo': ('Planet Albedo', None),
    #'calcTemperature()': ('Mean Planet Temperature (Calculated)', aq.K),
    #'estimateMass()': ('Planet Mass Estimated From Radius', aq.M_j),
    # 'calcSMA()': ('Semi-Major Axis', aq.au),
    # 'calcSMAfromT()': ('Semi-Major Axis (calculated from T)', aq.au),
    # 'calcPeriod': ('Planet Radius', aq.R_j),
    # 'discoveryMethod': ('Planet Radius', aq.R_j),
    'e': ('Orbital Eccentricity', None),
    'discoveryYear': ('Discovery Year', None),
    # 'lastUpdate': ('Last Updated', Date),
    'age': ('Planet Age', aq.Gyear),
    # 'ra': ('RA', None),
    # 'dec': ('DEC', None),
    'd': ('Distance to System', aq.pc),
    'R': ('Planet Radius', aq.R_j),
    'T': ('Planet Temperature', aq.K),
    'M': ('Planet Mass', aq.M_j),
    'calcSurfaceGravity()': ('Planet Surface Gravity', aq.ms2),
    'calcLogg()': ('Planet logg', None),
    'calcDensity()': ('Planet Density', aq.gcm3),
    'i': ('Orbital Inclination', aq.deg),
    'P': ('Period', aq.day),
    'a': ('Semi-Major Axis', aq.au),
    # 'transittime': ('Transit Time', aq.JulianDay),
    # TODO Catalogue Unit issue with theese
    # 'periastron': ('Orbit Periastron', aq.deg),
    # 'longitude': ('Orbit Longitude', aq.deg),
    # 'ascendingnode': ('Orbit Ascending Node', aq.deg),
}

_starPars = {
    # paramKey    : (axis label,      unit)
    'magV': ('V Magnitude', None),
    'magB': ('B Magnitude', None),
    'magH': ('H Magnitude', None),
    'magI': ('I Magnitude', None),
    'magJ': ('J Magnitude', None),
    'magK': ('K Magnitude', None),
    'd': ('Distance to System', aq.pc),
    'calcLuminosity()': ('Stellar Luminosity', aq.L_s),
    #'calcTemperature()': ('Stellar Temperature (Calculated)', aq.K),
    'Z': ('Stellar Metallicity', None),
    # 'Spectral Type': ('Spectral Type', str),
    # 'estimateAbsoluteMagnitude': ('Stellar Age', aq.Gyear),
    # 'estimateDistance': ('Stellar Age', aq.Gyear),
    'age': ('Stellar Age', aq.Gyear),
    # 'ra': ('RA', None),
    # 'dec': ('DEC', None),
    'R': ('Stellar Radius', aq.R_s),
    'T': ('Stellar Temperature', aq.K),
    'M': ('Stellar Mass', aq.M_s),
    'calcSurfaceGravity()': ('Planet Surface Gravity', aq.ms2),
    'calcLogg()': ('Stellar logg', None),
    'calcDensity()': ('Stellar Density', aq.gcm3),
}

# TODO Binary and System support for plots


""" Handles database classes including search functions
"""

import re
import xml.etree.ElementTree as ET
import glob
import os.path
import io
import gzip
import requests

from .astroclasses import System, Binary, Star, Planet, Parameters, BinaryParameters, StarParameters, PlanetParameters

compactString = lambda string: string.replace(' ', '').replace('-', '').lower()


class OECDatabase(object):
    """ This Class Handles the OEC database including search functions.
    """

    def __init__(self, databaseLocation, stream=False):
        """ Holds the Open Exoplanet Catalogue database in python

        :param databaseLocation: file path to the Open Exoplanet Catalogue systems folder ie
            ~/git/open-exoplanet-catalogue-atmospheres/systems/
            get the catalogue from https://github.com/hannorein/open_exoplanet_catalogue
            OR the stream object (used by load_db_from_url)
        :param stream: if true treats the databaseLocation as a stream object
        """

        self._loadDatabase(databaseLocation, stream)
        self._planetSearchDict = self._generatePlanetSearchDict()

        self.systemDict = dict((system.name, system) for system in self.systems)
        self.binaryDict = dict((binary.name, binary) for binary in self.binaries)
        self.starDict = dict((star.name, star) for star in self.stars)
        self.planetDict = dict((planet.name, planet) for planet in self.planets)

    def __repr__(self):
        return 'OECDatabase({} Systems, {} Binaries, {} Stars, {} Planets)'.format(len(self.systems), len(self.binaries),
                                                                                   len(self.stars), len(self.planets))

    def searchPlanet(self, name):
        """ Searches the database for a planet. Input can be complete ie GJ1214b, alternate name variations or even
        just 1214.

        :param name: the name of the planet to search
        :return: dictionary of results as planetname -> planet object
        """

        searchName = compactString(name)
        returnDict = {}

        for altname, planetObj in self._planetSearchDict.iteritems():
            if re.search(searchName, altname):
                returnDict[planetObj.name] = planetObj

        if returnDict:
            if len(returnDict) == 1:
                return returnDict.values()[0]
            else:
                return returnDict.values()

        else:
            return False

    @property
    def transitingPlanets(self):
        """ Returns a list of transiting planet objects
        """

        transitingPlanets = []

        for planet in self.planets:
            try:
                if planet.isTransiting:
                    transitingPlanets.append(planet)
            except KeyError:  # No 'discoverymethod' tag - this also filters Solar System planets
                pass

        return transitingPlanets

    def _generatePlanetSearchDict(self):
        """ Generates a search dictionary for planets by taking all names and 'flattening' them to the most compact form
        (lowercase, no spaces and dashes)
        """

        planetNameDict = {}
        for planet in self.planets:

            name = planet.name
            altnames = planet.params['altnames']
            altnames.append(name)  # as we also want the default name to be searchable

            for altname in altnames:
                reducedname = compactString(altname)
                planetNameDict[reducedname] = planet

        return planetNameDict

    def _loadDatabase(self, databaseLocation, stream=False):
        """ Loads the database from a given file path in the class

        :param databaseLocation: the location on disk or the stream object
        :param stream: if true treats the databaseLocation as a stream object
        """

        # Initialise Database
        self.systems = []
        self.binaries = []
        self.stars = []
        self.planets = []

        if stream:
            tree = ET.parse(databaseLocation)
            for system in tree.findall(".//system"):
                self._loadSystem(system)
        else:
            databaseXML = glob.glob(os.path.join(databaseLocation, '*.xml'))
            if not len(databaseXML):
                raise LoadDataBaseError('could not find the database xml files. Have you given the correct location '
                                        'to the open exoplanet catalogues /systems folder?')

            for filename in databaseXML:
                try:
                    with open(filename, 'r') as f:
                        tree = ET.parse(f)
                except ET.ParseError as e:  # this is sometimes raised rather than the root.tag system check
                    raise LoadDataBaseError(e)

                root = tree.getroot()

                # Process the system
                if not root.tag == 'system':
                    raise LoadDataBaseError('file {0} does not contain a valid system - could be an error with your version'
                                            ' of the catalogue'.format(filename))

                self._loadSystem(root)

    def _loadSystem(self, root):
        systemParams = Parameters()
        for systemXML in root:

            tag = systemXML.tag
            text = systemXML.text
            attrib = systemXML.attrib

            systemParams.addParam(tag, text, attrib)

        system = System(systemParams.params)
        self.systems.append(system)  # Add system to the index

        self._loadBinarys(root, system)
        self._loadStars(root, system)

    def _loadBinarys(self, parentXML, parent):

        binarysXML = parentXML.findall("binary")

        for binaryXML in binarysXML:
            binaryParams = BinaryParameters()

            for value in binaryXML:

                tag = value.tag
                text = value.text
                attrib = value.attrib

                binaryParams.addParam(tag, text, attrib)

            binary = Binary(binaryParams.params)
            binary.parent = parent

            parent._addChild(binary)  # Add star to the system

            self._loadBinarys(binaryXML, binary)
            self._loadStars(binaryXML, binary)
            self._loadPlanets(binaryXML, binary)  # Load planets

            self.binaries.append(binary)  # Add star to the index

    def _loadStars(self, parentXML, parent):

        starsXML = parentXML.findall("star")

        for starXML in starsXML:
            starParams = StarParameters()

            for value in starXML:

                tag = value.tag
                text = value.text
                attrib = value.attrib

                starParams.addParam(tag, text, attrib)

            star = Star(starParams.params)
            star.parent = parent

            parent._addChild(star)  # Add star to the system

            self._loadPlanets(starXML, star)  # Load planets

            self.stars.append(star)  # Add star to the index

    def _loadPlanets(self, parentXML, parent):

        planetsXML = parentXML.findall("planet")

        for planetXML in planetsXML:

            planetParams = PlanetParameters()

            for value in planetXML:

                tag = value.tag
                text = value.text
                attrib = value. attrib

                planetParams.addParam(tag, text, attrib)

            planet = Planet(planetParams.params)
            planet.parent = parent

            parent._addChild(planet)  # Add planet to the star
            self.planets.append(planet)  # Add planet to the index


def load_db_from_url(url="https://github.com/OpenExoplanetCatalogue/oec_gzip/raw/master/systems.xml.gz"):
    """ Loads the database from a gzipped version of the system folder, by default the one located in the oec_gzip repo
    in the OpenExoplanetCatalogue GitHub group.

    The database is loaded from the url in memory

    :param url: url to load (must be gzipped version of systems folder)
    :return: OECDatabase objected initialised with latest OEC Version
    """

    catalogue = gzip.GzipFile(fileobj=io.BytesIO(requests.get(url).content))
    database = OECDatabase(catalogue, stream=True)

    return database


class LoadDataBaseError(IOError):
    pass


"""
Exoplanet related equatio. Each equation is a class which takes input of every
variable bar one and will output the variable left out. They therefore implement
all permutations of the equation with respect to the variables.

As an example we choose the equation Kepler's Third Law (KeplersThirdLaw) and
by defining the values for GJ 1214 b we can calculate the period.

    >>> from exodata.equations import KeplersThirdLaw
    >>> KeplersThirdLaw(a=0.01488*aq.au, M_s=0.176*aq.M_s).P
    array(1.579696141940911) * d

Equations are named based on their main purpose or common name. I.e the equation
scale height is :math:`H = \frac{k T_{eff}}{\mu g}` even though we could use it
to calculate g, given the other parameters.

Equations are designed to be user friendly and accurate --- not fast. This means
if you are using this as part of a large simulation and use ExoData to generate
initial parameters you'll be fine. If however you are running a certain equation
millions of times it may be worth using your own optimised version.

**Abbreviations used in this module**

Where abbreviations are ambiguous (i.e. *R* could be the radius of anything)
we use the suffix *_p* for the planet and *_s* for the star. A list of
parameters used is given below.

* *R_p* - Planetary Radius
* *M_p* - Planetary Mass
* *M_s* - Stellar Mass
* *R_s* - Stellar Radius
* *H* - Scale height
* *i* - orbital inclination
* *e* - orbit eccentricity
* *T_eff_s* - Effective Temperature of the Star
* *A* - Albedo
* *mu* - mean molecular weight
"""

from __future__ import division
from numpy import sqrt, arcsin, sin, cos, log10, nan

import numpy as np
import os
import sys
from pkg_resources import resource_stream, resource_filename
import math

import quantities.constants as const
from . import astroquantities as aq
from . import params


pi = const.pi
sigma = const.Stefan_Boltzmann_constant
G = const.Newtonian_constant_of_gravitation

_rootdir = os.path.dirname(__file__)


class _ExoDataEqn(object):

    def __init__(self):
        self.vars = (None,)

    def __repr__(self):
        vs = ['{}={}'.format(v, eval('self._{}'.format(v)), self)
              for v in self.vars if v is not None]
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(vs))  # skip final ', '


class ScaleHeight(_ExoDataEqn):

    def __init__(self, T_eff=None, mu=None, g=None, H=None):
        r""" Uses the scale height equation to calculate a parameter given the
        others.

        .. math::
            H = \frac{k T_{eff}{\mu g}

        Where H is the scale height of the planets atmosphere, :math:`T_{eff}`
        is the planetary effective temperature,
        :math:`\mu` is the mean molecular weight of the planetary atmosphere and
        g is the planets surface gravity.

        :param T_eff: Effective temperature of the planet
        :param mu: mean molecular weight for the atmosphere
        :param g: surface gravity of the planet
        :return: H (scale height) of the atmosphere
        """

        _ExoDataEqn.__init__(self)

        self._T_eff = T_eff
        self._mu = mu
        self._g = g
        self._H = H

        self.vars = ('H', 'T_eff', 'mu', 'g')  # list of input variables

        if (T_eff, mu, g, H).count(None) > 1:
            raise EqnInputError("You must give all parameters bar one")

    @property
    def H(self):

        H = self._H
        mu = self._mu
        g = self._g
        T_eff = self._T_eff

        if H is None:
            H = (const.k * T_eff) / (mu * g)

        return H.rescale(aq.m)

    @property
    def T_eff(self):

        H = self._H
        mu = self._mu
        g = self._g
        T_eff = self._T_eff

        if T_eff is None:
            T_eff = (H * mu * g) / const.k

        return T_eff.rescale(aq.K)

    @property
    def mu(self):

        H = self._H
        mu = self._mu
        g = self._g
        T_eff = self._T_eff

        if mu is None:
            mu = (const.k * T_eff) / (H * g)

        return mu.rescale(aq.atomic_mass_unit)

    @property
    def g(self):

        H = self._H
        mu = self._mu
        g = self._g
        T_eff = self._T_eff

        if g is None:
            g = (const.k * T_eff) / (H * mu)

        return g.rescale(aq.m / aq.s**2)


class MeanPlanetTemp(_ExoDataEqn):

    def __init__(self, A, T_s, R_s, a, epsilon=0.7, T_p=None):
        """ Calculate the equilibrium planet temperature

        .. math::

        assumes epsilon = 0.7 by default http://arxiv.org/pdf/1111.1455v2.pdf
        """
        _ExoDataEqn.__init__(self)

        self._A = A
        self._T_s = T_s
        self._T_p = T_p
        self._R_s = R_s
        self._a = a
        self._epsilon = epsilon

        self.vars = ('A', 'T_s', 'T_p', 'R_s', 'a',
                     'epsilon')  # list of input variables

        if (A, T_s, R_s, a, epsilon, T_p).count(None) > 1:
            raise EqnInputError("You must give all parameters bar one")

    @property
    def T_p(self):

        T_p = self._T_p
        A = self._A
        T_s = self._T_s
        R_s = self._R_s
        a = self._a
        epsilon = self._epsilon

        if T_p is None:
            T_p = T_s * ((1 - A) / epsilon)**(1 / 4) * sqrt(R_s / (2 * a))

        return T_p.rescale(aq.degK)

    @property
    def A(self):

        T_p = self._T_p
        A = self._A
        T_s = self._T_s
        R_s = self._R_s
        a = self._a
        epsilon = self._epsilon

        if A is None:
            A = 1 - epsilon * (T_p / (T_s * sqrt(R_s / (2 * a))))**4

        return A.rescale(aq.dimensionless)

    @property
    def T_s(self):

        T_p = self._T_p
        A = self._A
        T_s = self._T_s
        R_s = self._R_s
        a = self._a
        epsilon = self._epsilon

        if T_s is None:
            T_s = T_p / (((1 - A) / epsilon)**(1 / 4) * sqrt(R_s / (2 * a)))

        return T_s.rescale(aq.K)

    @property
    def R_s(self):

        T_p = self._T_p
        A = self._A
        T_s = self._T_s
        R_s = self._R_s
        a = self._a
        epsilon = self._epsilon

        if R_s is None:
            R_s = 2 * a * (T_p / (T_s * ((1 - A) / epsilon)**(1 / 4)))**2

        return R_s.rescale(aq.R_s)

    @property
    def a(self):

        T_p = self._T_p
        A = self._A
        T_s = self._T_s
        R_s = self._R_s
        a = self._a
        epsilon = self._epsilon

        if a is None:
            a = R_s / (2 * (T_p / (T_s * ((1 - A) / epsilon)**(1 / 4)))**2)

        return a.rescale(aq.au)

    @property
    def epsilon(self):

        T_p = self._T_p
        A = self._A
        T_s = self._T_s
        R_s = self._R_s
        a = self._a
        epsilon = self._epsilon

        if epsilon is None:
            epsilon = (1 - A) / (T_p / (T_s * sqrt(R_s / (2 * a)))) ** 4

        return epsilon.rescale(aq.dimensionless)


class StellarLuminosity(_ExoDataEqn):

    def __init__(self, R=None, T=None, L=None):
        r""" Calculate stellar luminosity

        .. math::
            L_\star = 4\pi R^2_\star \sigma T^4_\star

        Where :math:`L_\star` is the Stellar luminosity, :math:`R_\star` stellar
        radius, :math:`\sigma` Stefan-Boltzman constant, :math:`T_\star` the
        temperature of the star

        :param R: stellar radius
        :param T: effective surface temperature of the star
        :param L: Stellar luminosity
        """

        _ExoDataEqn.__init__(self)

        self._L = L
        self._R = R
        self._T = T

        self.vars = ('R', 'T', 'L')  # list of input variables

        if (R, T, L).count(None) > 1:
            raise EqnInputError("You must give all parameters bar one")

    @property
    def L(self):

        L = self._L
        R = self._R
        T = self._T

        if L is None:
            L = 4 * pi * R**2 * sigma * T**4

        return L.rescale(aq.W)

    @property
    def R(self):

        L = self._L
        R = self._R
        T = self._T

        if R is None:
            R = sqrt(L / (4 * pi * sigma * T**4))

        return R.rescale(aq.R_s)

    @property
    def T(self):

        L = self._L
        R = self._R
        T = self._T

        if T is None:
            T = (L / (4 * pi * sigma * R**2))**0.25

        return T.rescale(aq.K)


class KeplersThirdLaw(_ExoDataEqn):

    def __init__(self, a=None, M_s=None, P=None, M_p=0. * aq.M_j):
        r"""Calculates the period of the orbit using the stellar mass, planet
        mass and sma using keplers Third Law

        .. math::
            P = \sqrt{\frac{4\pi^2a^3}{G \left(M_\star + M_p \right)}}
        """

        _ExoDataEqn.__init__(self)

        self._a = a
        self._M_s = M_s
        self._P = P
        self._M_p = M_p

        if (a, M_s, P).count(None) > 1:
            raise EqnInputError("You must give all parameters bar one")

    @property
    def P(self):

        a = self._a
        M_s = self._M_s
        M_p = self._M_p
        P = self._P

        if P is None:
            P = 2 * pi * sqrt(a**3 / (G * (M_s + M_p)))

        return P.rescale(aq.day)

    @property
    def a(self):

        a = self._a
        M_s = self._M_s
        M_p = self._M_p
        P = self._P

        try:
            if a is None:
                a = ((P**2 * G * (M_s + M_p)) / (4 * pi**2))**(1. / 3)
            return a.rescale(aq.au)
        except ValueError:
            return np.nan

    @property
    def M_s(self):

        a = self._a
        M_s = self._M_s
        M_p = self._M_p
        P = self._P

        if M_s is None:
            M_s = ((4 * pi**2 * a**3) / (G * P**2)) - M_p

        return M_s.rescale(aq.M_s)

    @property
    def M_p(self):

        a = self._a
        M_s = self._M_s
        M_p = self._M_p
        P = self._P

        if M_p is None:
            M_p = ((4 * pi**2 * a**3) / (G * P**2)) - M_s

        return M_p.rescale(aq.M_j)


class SurfaceGravity(_ExoDataEqn):

    def __init__(self, M=None, R=None, g=None):
        r""" Calculates the surface acceleration due to gravity on the planet

        .. math::
            g_p = \frac{GM_p}{R_p^2}

        where :math:`g_p` is the acceleration ude to gravity on the planet
        surface, G is the gravitational constant,
        :math:`M_p` planetary mass, :math:`R_p` radius of the planet

        :param params: dict containing 'M_p', 'R_p' with units

        :return: g - acceleration due to gravity * m / s**2
        """

        _ExoDataEqn.__init__(self)

        self._M = M
        self._R = R
        self._g = g

        if (M, R, g).count(None) > 1:
            raise EqnInputError("You must give all parameters bar one")

    @property
    def g(self):

        M = self._M
        R = self._R
        g = self._g

        if g is None:
            g = (G * M) / (R**2)

        return g.rescale(aq.m / aq.s**2)

    @property
    def M(self):

        M = self._M
        R = self._R
        g = self._g

        if M is None:
            M = (g * R**2) / G

        return M.rescale(aq.M_j)

    @property
    def R(self):

        M = self._M
        R = self._R
        g = self._g

        if R is None:
            R = sqrt(M * G / g)

        return R.rescale(aq.R_j)


class Logg(_ExoDataEqn):

    def __init__(self, M=None, R=None, logg=None):
        """Calculates the surface acceleration due to gravity on the planet as
        logg, the base 10 logarithm of g in cgs units. This function uses
        :py:func:`surfaceGravity` and then rescales it to cgs and takes the log.
        """

        _ExoDataEqn.__init__(self)

        self._M = M
        self._R = R
        self._logg = logg

        if (M, R, logg).count(None) > 1:
            raise EqnInputError("You must give all parameters bar one")

    @property
    def logg(self):

        M = self._M
        R = self._R
        logg = self._logg

        if logg is None:
            g = SurfaceGravity(M, R).g
            # the float wrapper is needed to remove dimensionality
            logg = log10(float(g.rescale(aq.cm / aq.s**2)))

        return logg

    @property
    def M(self):

        M = self._M
        R = self._R
        logg = self._logg

        if M is None:
            g = (10**logg) * aq.cm / aq.s**2
            M = SurfaceGravity(None, R, g).M

        return M.rescale(aq.M_j)

    @property
    def R(self):

        M = self._M
        R = self._R
        logg = self._logg

        if R is None:
            g = (10**logg) * aq.cm / aq.s**2
            R = SurfaceGravity(M, None, g).R

        return R.rescale(aq.R_j)


class TransitDepth(_ExoDataEqn):

    def __init__(self, R_s=None, R_p=None, depth=None):

        _ExoDataEqn.__init__(self)

        self._R_s = R_s
        self._R_p = R_p
        self._depth = depth

        if (R_s, R_p, depth).count(None) > 1:
            raise EqnInputError("You must give all parameters bar one")

    @property
    def depth(self):

        R_s = self._R_s
        R_p = self._R_p
        depth = self._depth

        if depth is None:
            depth = (R_p / R_s)**2

        return depth.rescale(aq.dimensionless)

    @property
    def R_s(self):

        R_s = self._R_s
        R_p = self._R_p
        depth = self._depth

        if R_s is None:
            R_s = R_p / sqrt(depth)

        return R_s

    @property
    def R_p(self):

        R_s = self._R_s
        R_p = self._R_p
        depth = self._depth

        if R_p is None:
            R_p = R_s * sqrt(depth)

        return R_p.rescale(aq.R_j)


class Density(_ExoDataEqn):

    def __init__(self, M=None, R=None, density=None):
        """ Calculates the density in g/cm**3
        :param R: radius
        :param M: mass
        :return:
        """

        _ExoDataEqn.__init__(self)

        self._M = M
        self._R = R
        self._density = density

        if (M, R, density).count(None) > 1:
            raise EqnInputError("You must give all parameters bar one")

    @property
    def density(self):

        M = self._M
        R = self._R
        density = self._density

        if density is None:
            volume = 4. / 3 * pi * R**3
            density = (M / volume)

        return density.rescale(aq.g / aq.cm**3)

    @property
    def M(self):

        M = self._M
        R = self._R
        density = self._density

        if M is None:
            volume = 4. / 3 * pi * R**3
            M = density * volume

        return M.rescale(aq.M_j)

    @property
    def R(self):

        M = self._M
        R = self._R
        density = self._density

        if R is None:
            R = (M / (density * 4. / 3 * pi))**(1. / 3)

        return R.rescale(aq.R_j)


class TransitDuration(_ExoDataEqn):

    def __init__(self, P=None, a=None, Rp=None, Rs=None, i=None, e=None, w=None):
        r"""Eccentric transit duration equation from Kipping (2011).

        Currently only calculates TD given the other values. There is little
        demand for the rearrangements and they are non trivial to implement.

        .. math::
            T_{14} = \frac{P}{\pi} \frac{\varrho_{P,T}^2}{\sqrt{1-e^2}}
            \arcsin{\left( \sqrt{\frac{S^2_{P*} -
            b^2_{P,T}}{(a_P/R_\star)^2\varrho^2_{P,T} -b^2_{P,T}}}\right)}

            a_R = (a / R_\star)

            \varrho_{P,T}(f_p) = \frac{1-e^2_P}{1+e_P \sin(\omega)}

            b_{P,T} = (a_P / R_\star)\varrho_{P,T}\cos{i}
        """

        _ExoDataEqn.__init__(self)

        self.P = P
        self.a = a
        self.Rp = Rp
        self.Rs = Rs
        self.i = i
        self.e = e
        Td = None

        if w == 0:
            self.w = 0 * aq.rad
        else:
            self.w = w

        if (P, a, Rp, Rs, i, e, w, Td).count(None) > 1:
            raise EqnInputError("You must give all parameters bar one")

    @property
    def Td(self):

        a = (self.a / self.Rs).rescale(aq.dimensionless)
        i = self.i.rescale(aq.rad)
        w = self.w.rescale(aq.rad)
        P = self.P
        RpRs = (self.Rp / self.Rs).rescale(aq.dimensionless)
        e = self.e

        ro_pt = (1 - e**2) / (1 + e * np.sin(w))
        b_pt = a * ro_pt * np.cos(i)
        s_ps = 1.0 + RpRs
        df = np.arcsin(np.sqrt((s_ps**2 - b_pt**2) /
                               ((a**2) * (ro_pt**2) - b_pt**2)))

        duration = (P * (ro_pt**2)) / (np.pi * np.sqrt(1 - e**2)) * df

        return duration.rescale(aq.min)


class ImpactParameter(_ExoDataEqn):

    def __init__(self, a=None, R_s=None, i=None, b=None):
        r"""Projected distance between the planet and star centres during
        mid-transit.

        .. math::
            b \equiv \frac{a}{R_*} \cos{i}

        (Seager & Mallen-Ornelas 2003).
        """

        _ExoDataEqn.__init__(self)

        self._a = a
        self._R_s = R_s
        self._i = i
        self._b = b

        if (a, R_s, i, b).count(None) > 1:
            raise EqnInputError("You must give all parameters bar one")

    @property
    def b(self):

        a = self._a
        R_s = self._R_s
        i = self._i

        b = (a / R_s) * cos(i.rescale(aq.rad))

        return b.rescale(aq.dimensionless)

    @property
    def a(self):

        b = self._b
        R_s = self._R_s
        i = self._i

        a = (b * R_s) / cos(i.rescale(aq.rad))

        return a.rescale(aq.au)

    @property
    def R_s(self):

        a = self._a
        b = self._b
        i = self._i

        R_s = (a / b) * cos(i.rescale(aq.rad))

        return R_s.rescale(aq.R_s)

    @property
    def i(self):

        a = self._a
        b = self._b
        R_s = self._R_s

        i = np.arccos((R_s * b / a).rescale(aq.dimensionless))

        return i.rescale(aq.deg)


def ratioTerminatorToStar(H_p, R_p, R_s):  # TODO add into planet class
    r"""Calculates the ratio of the terminator to the star assuming 5 scale
    heights large. If you dont know all of the input try
    :py:func:`calcRatioTerminatorToStar`

    .. math::
        \Delta F = \frac{10 H R_p + 25 H^2}{R_\star^2}

    Where :math:`\Delta F` is the ration of the terminator to the star,
    H scale height planet atmosphere, :math:`R_p` radius of the planet,
    :math:`R_s` radius of the star

    :param H_p:
    :param R_p:
    :param R_s:
    :return: ratio of the terminator to the star
    """

    deltaF = ((10 * H_p * R_p) + (25 * H_p**2)) / (R_s**2)
    return deltaF.simplified


def SNRPlanet(SNRStar, starPlanetFlux, Nobs, pixPerbin, NVisits=1):
    r""" Calculate the Signal to Noise Ratio of the planet atmosphere

    .. math::
        \text{SNR}_\text{planet} = \text{SNR}_\text{star} \times \Delta F \times
        \sqrt{N_\text{obs}}
        \times \sqrt{N_\text{pixPerbin}} \times \sqrt{N_\text{visits}}

    Where :math:`\text{SNR}_\star` SNR of the star detection, :math:`\Delta F`
    ratio of the terminator to the star, :math:`N_\text{obs}` number of
    exposures per visit, :math:`N_\text{pixPerBin}` number of pixels per
    wavelength bin, :math:`N_\text{visits}` number of visits.

    :return:
    """

    SNRplanet = SNRStar * starPlanetFlux * \
        sqrt(Nobs) * sqrt(pixPerbin) * sqrt(NVisits)

    return SNRplanet


def transitDurationCircular(P, R_s, R_p, a, i):
    r"""Estimation of the primary transit time. Assumes a circular orbit.

    .. math::
        T_\text{dur} = \frac{P}{\pi}\sin^{-1}
        \left[\frac{R_\star}{a}\frac{\sqrt{(1+k)^2 + b^2}}{\sin{a}} \right]

    Where :math:`T_\text{dur}` transit duration, P orbital period,
    :math:`R_\star` radius of the star, a is the semi-major axis,
    k is :math:`\frac{R_p}{R_s}`, b is :math:`\frac{a}{R_*} \cos{i}`

    (Seager & Mallen-Ornelas 2003)
    """

    if i is nan:
        i = 90 * aq.deg

    i = i.rescale(aq.rad)
    k = R_p / R_s  # lit reference for eclipsing binaries
    b = (a * cos(i)) / R_s

    duration = (P / pi) * arcsin(((R_s * sqrt((1 + k) **
                                              2 - b ** 2)) / (a * sin(i))).simplified)

    return duration.rescale(aq.min)


def estimateStellarTemperature(M_s):
    """Estimates stellar temperature using the main sequence relationship
    T ~ 5800*M^0.65 (Cox 2000).
    """
    # TODO improve with more x and k values from Cox 2000
    try:
        temp = (5800 * aq.K * float(M_s.rescale(aq.M_s) ** 0.65)).rescale(aq.K)
    except AttributeError:
        temp = np.nan
    return temp


def estimateDistance(m, M, Av=0.0):
    """ estimate the distance to star based on the absolute magnitude, apparent
    magnitude and the absorption / extinction.

    :param m: apparent magnitude
    :param M: absolute magnitude
    :param Av: absorbtion / extinction

    :return: d (distance to object) in parsecs
    """
    try:
        m = float(m)  # basic value checking as there is no units
        M = float(M)
        Av = float(Av)
    except TypeError:
        return np.nan

    d = 10 ** ((m - M + 5 - Av) / 5)

    if math.isnan(d):
        return np.nan
    else:
        return d * aq.pc


def _createAbsMagEstimationDict():
    """ loads magnitude_estimation.dat which is from
    http://xoomer.virgilio.it/hrtrace/Sk.htm on 24/01/2014 and based on
    Schmid-Kaler (1982)

    creates a dict in the form [Classletter][ClassNumber][List of values for
    each L Class]
    """
    magnitude_estimation_filepath = resource_filename(
        __name__, 'data/magnitude_estimation.dat')
    raw_table = np.loadtxt(magnitude_estimation_filepath, '|S5')

    absMagDict = {
        'O': {},
        'B': {},
        'A': {},
        'F': {},
        'G': {},
        'K': {},
        'M': {}}
    for row in raw_table:
        if sys.hexversion >= 0x03000000:
            # otherwise we get byte ints or b' caused by 2to3
            starClass = row[0].decode("utf-8")
            absMagDict[starClass[0]][int(starClass[1])] = [
                float(x) for x in row[1:]]
        else:
            # dict of spectral type = {abs mag for each luminosity class}
            absMagDict[row[0][0]][int(row[0][1])] = [float(x) for x in row[1:]]

    # manually typed from table headers - used to match columns with the L
    # class (header)
    LClassRef = {
        'V': 0,
        'IV': 1,
        'III': 2,
        'II': 3,
        'Ib': 4,
        'Iab': 5,
        'Ia': 6,
        'Ia0': 7}

    return absMagDict, LClassRef

absMagDict, LClassRef = _createAbsMagEstimationDict()


def estimateAbsoluteMagnitude(spectralType):
    """Uses the spectral type to lookup an approximate absolute magnitude for
    the star.
    """

    from .astroclasses import SpectralType

    specType = SpectralType(spectralType)

    if specType.classLetter == '':
        return np.nan
    elif specType.classNumber == '':
        specType.classNumber = 5  # approximation using mid magnitude value

    if specType.lumType == '':
        specType.lumType = 'V'  # assume main sequence

    LNum = LClassRef[specType.lumType]
    classNum = specType.classNumber
    classLet = specType.classLetter

    try:
        return absMagDict[classLet][classNum][LNum]
    # value not in table. Assume the number isn't there (Key p2.7, Ind p3+)
    except (KeyError, IndexError):
        try:
            classLookup = absMagDict[classLet]
            values = np.array(list(classLookup.values()))[
                :, LNum]  # only select the right L Type
            return np.interp(classNum, list(classLookup.keys()), values)
        except (KeyError, ValueError):
            return np.nan  # class not covered in table


def _createMagConversionDict():
    """Loads magnitude_conversion.dat which is table A% 1995ApJS..101..117K
    """
    magnitude_conversion_filepath = resource_stream(
        __name__, 'data/magnitude_conversion.dat')
    raw_table = np.loadtxt(magnitude_conversion_filepath, '|S5')

    magDict = {}
    for row in raw_table:
        if sys.hexversion >= 0x03000000:
            # otherwise we get byte ints or b' caused by 2to3
            starClass = row[1].decode("utf-8")
            tableData = [x.decode("utf-8") for x in row[3:]]
        else:
            starClass = row[1]
            tableData = row[3:]
        magDict[starClass] = tableData

    return magDict

magDict = _createMagConversionDict()


def magKtoMagV(*args, **kwargs):
    """ Converts K magnitude to V magnitude."""
    raise DeprecationWarning(
        "This function has is been phased out in favour of the"
        "astroclasses.Magnitude class which can convert between many magnitudes")


class EqnInputError(params.ExoDataError):
    pass
